#include "cuda_check.cuh"
#include <math.h>
#include "timer.hxx"
#include "produit_matrices.hxx"

void alloc_CPU(hMatrice & h_u, hMatrice & h_v, hMatrice & h_w, hMatrice & h_verif, int n, int m, int p)
{
  Timer T("Allocation sur CPU");

  h_u.allocate(n, m);
  h_v.allocate(m, p);
  h_w.allocate(n, p);
  h_verif.allocate(n, p);
  
  std::cout << T << std::endl;
}

void alloc_GPU(dMatrice &d_u, dMatrice &d_v, dMatrice &d_w, int n, int m, int p)
{
  Timer T("Allocation sur GPU");
  
  d_u.allocate(n, m);
  d_v.allocate(m, p);
  d_w.allocate(n, p);

  std::cout << T << std::endl;
}

void init_CPU(hMatrice &h_u, hMatrice &h_v)
{
  Timer T("Initialisation sur CPU");
  int i,j, n = h_u.n(), m = h_u.m(), p = h_v.m();

  for (i=0; i<n; i++)
    for (j=0; j<m; j++)
      h_u(i,j) = (2.0*i+j+1)/(n*m);

  float *c_v = h_v.data();
  for (i=0; i<m; i++)
    for (j=0; j<p; j++)
      h_v(i,j) = (2.0*i-j-1)/(n*m);

  std::cout << T << std::endl;
}

void matMul_CPU(hMatrice &h_c, hMatrice &h_a, hMatrice &h_b)
{
  Timer T("Calcul sur CPU", 1);
  int i, n = h_a.n(), m = h_a.m(), p = h_b.m();

  float * a = h_a.data();
  float * b = h_b.data();
  float * c = h_c.data();
  
  size_t idx, idy, ida, idb, idc;

  for (idx=0; idx<n; idx++)
    for (idy=0; idy<p; idy++) {

      double s = 0;
      ida = idy*n;
      idb = idx;
      for (i=0; i<m; i++) {
        s += a[ida] * b[idb];
        ida += 1;
        idb += m;
      }
      idc = idx + idy * n;
      c[idc] = s;
    }
  std::cout << "\n" << T << "\n" << std::endl;
}

 
void copy_CPU_GPU(dMatrice & d_u, dMatrice &d_v, hMatrice & h_u, hMatrice & h_v)
{
  Timer T("Copie CPU -> GPU");
  size_t nb = d_u.bytes();

  cudaMemcpy(d_u.data(), h_u.data(), nb, cudaMemcpyHostToDevice);
  cudaMemcpy(d_v.data(), h_v.data(), nb, cudaMemcpyHostToDevice);

  std::cout << T << std::endl;
}

__global__ void _matMul_GPU(float *c, float *a, float *b, size_t n, size_t m, size_t p)
{
  size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  size_t idy = blockIdx.y*blockDim.y+threadIdx.y;
  size_t i, ida, idb, idc;

  if (idx >= m || idy >= p) return;
  
  double s = 0;
  ida = idy * n;
  idb = idx;
  for (i=0; i<m; i++)
  {
    s += a[ida] * b[idb];
    ida += 1;
    idb += p;
  }
  idc = idx + idy * m;
  c[idc] = s;
}

void matMul_GPU(dMatrice &d_w, dMatrice &d_u, dMatrice &d_v)
{
  Timer T("Calcul sur GPU", 1);

  int g = 16;
  int d_n = ceil(((double) d_w.n())/16);
  int d_p = ceil(((double) d_w.m())/16);

  dim3 blockSize(g,g,1);
  dim3 gridSize(d_n,d_p,1);

  _matMul_GPU<<<gridSize, blockSize>>>(d_w.data(), d_u.data(), d_v.data(), d_u.n(), d_u.m(), d_w.m());

  cudaDeviceSynchronize();
  CUDA_CHECK_KERNEL();

  std::cout << "\n" << T << "\n" << std::endl;
}

void copy_GPU_CPU(hMatrice &h_w, dMatrice &d_w)
{
  Timer T("Copie GPU -> CPU");

  cudaMemcpy(h_w.data(), d_w.data(), d_w.bytes(), cudaMemcpyDeviceToHost);

  std::cout << T << std::endl;
}

double verification(hMatrice &h_w, hMatrice &h_verif)
{
  Timer T("Verification CPU - GPU");

  double s = 0.0, v = 0.0;
  size_t nm = h_verif.n()*h_verif.m();
  float * c_verif = h_verif.data();
  float * c_w = h_w.data();

  for(size_t k=0; k<nm; k++) {
    s += fabs(c_verif[k] - c_w[k]);
    v += fabs(c_verif[k] + c_w[k]);
  }

  std::cout << "\n" << T << std::endl;

  return s/v;
}

#include <cublas_v2.h>

void matMul_GPU_Cublas(dMatrice &d_w, dMatrice &d_u, dMatrice &d_v)
{
  Timer T("Calcul sur GPU (Cublas)", 1);

  cublasHandle_t handle;
  cublasCreate(&handle);

  float alpha = 1.0, beta = 0.0;

  cublasSgemm(handle, 
        CUBLAS_OP_N, CUBLAS_OP_N, 
        d_v.m(), d_u.n(), d_u.m(),
        &alpha, d_v.data(), d_v.m(), 
                d_u.data(), d_u.m(), 
        &beta,  d_w.data(), d_w.m());

  std::cout << "\n" << T << "\n" << std::endl;
}

